@@ -2,7 +2,7 @@
  * Create a squashfs filesystem.  This is a highly compressed read only filesystem.
  *
  * Copyright (c) 2002, 2003, 2004, 2005
- * Phillip Lougher <phillip@lougher.demon.co.uk>
+ * Phillip Lougher <phillip@lougher.org.uk>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -36,20 +36,29 @@
 
 #include <squashfs_fs.h>
 #include "global.h"
+#include "sort.h"
 
 #ifdef SQUASHFS_TRACE
-#define TRACE(s, args...)		printf("mksquashfs: "s, ## args)
+#define TRACE(s, args...)		do { \
+						printf("mksquashfs: "s, ## args); \
+					} while(0)
 #else
 #define TRACE(s, args...)
 #endif
 
-#define INFO(s, args...)		if(!silent) printf("mksquashfs: "s, ## args)
-#define ERROR(s, args...)		fprintf(stderr, s, ## args)
-#define EXIT_MKSQUASHFS()		exit(1)
-#define BAD_ERROR(s, args...)		{\
-					fprintf(stderr, "FATAL ERROR:" s, ##args);\
-					EXIT_MKSQUASHFS();\
-					}
+#define INFO(s, args...)		do { \
+						if(!silent) printf("mksquashfs: "s, ## args); \
+					} while(0)
+#define ERROR(s, args...)		do { \
+						fprintf(stderr, s, ## args); \
+					} while(0)
+#define EXIT_MKSQUASHFS()		do { \
+						exit(1); \
+					} while(0)
+#define BAD_ERROR(s, args...)		do {\
+						fprintf(stderr, "FATAL ERROR:" s, ##args);\
+						EXIT_MKSQUASHFS();\
+					} while(0);
 
 int mkisofs_style = -1;
 
@@ -63,67 +72,17 @@ struct sort_info {
 struct sort_info *sort_info_list[65536];
 
 struct priority_entry {
-	char *filename;
-	long long size;
-	ino_t st_ino;
-	dev_t st_dev;
+	struct dir_ent *dir;
 	struct priority_entry *next;
 };
 
 struct priority_entry *priority_list[65536];
 
-struct sorted_inode_entry {
-	squashfs_inode inode;
-	ino_t	st_ino;
-	dev_t	st_dev;
-	struct sorted_inode_entry *next;
-};
-
-struct sorted_inode_entry *sorted_inode_list[65536];
-
 extern int silent;
-extern int excluded(char *filename, struct stat *buf);
-extern squashfs_inode write_file(squashfs_inode *inode, char *filename, long long size, int *c_size);
+extern squashfs_inode write_file(squashfs_inode *inode, struct dir_ent *dir_ent, long long size, int *c_size);
 
 
-int add_to_sorted_inode_list(squashfs_inode inode, dev_t st_dev, ino_t st_ino)
-{
-	int hash = st_ino & 0xffff;
-	struct sorted_inode_entry *new_sorted_inode_entry;
-
-	if((new_sorted_inode_entry = malloc(sizeof(struct sorted_inode_entry))) == NULL) {
-		ERROR("Out of memory allocating sorted inode entry\n");
-		return FALSE;
-	}
-
-	new_sorted_inode_entry->inode = inode;
-	new_sorted_inode_entry->st_ino = st_ino;
-	new_sorted_inode_entry->st_dev = st_dev;
-	new_sorted_inode_entry->next = sorted_inode_list[hash];
-	sorted_inode_list[hash] = new_sorted_inode_entry;
-
-	return TRUE;
-}
-
-	
-int get_sorted_inode(squashfs_inode *inode, struct stat *buf)
-{
-	int hash = buf->st_ino & 0xffff;
-	struct sorted_inode_entry *sorted_inode_entry;
-
-	for(sorted_inode_entry = sorted_inode_list[hash]; sorted_inode_entry; sorted_inode_entry = sorted_inode_entry->next)
-		if(buf->st_ino == sorted_inode_entry->st_ino && buf->st_dev == sorted_inode_entry->st_dev)
-			break;
-
-	if(sorted_inode_entry) {
-		*inode = sorted_inode_entry->inode;
-		return 1;
-	} else
-		return 0;
-}
-
-
-int add_priority_list(char *filename, struct stat *buf, int priority)
+int add_priority_list(struct dir_ent *dir, int priority)
 {
 	struct priority_entry *new_priority_entry;
 
@@ -133,10 +92,7 @@ int add_priority_list(char *filename, struct stat *buf, int priority)
 		return FALSE;
 	}
 
-	new_priority_entry->filename = strdup(filename);
-	new_priority_entry->size = buf->st_size;
-	new_priority_entry->st_dev = buf->st_dev;
-	new_priority_entry->st_ino = buf->st_ino;
+	new_priority_entry->dir = dir;;
 	new_priority_entry->next = priority_list[priority];
 	priority_list[priority] = new_priority_entry;
 	return TRUE;
@@ -222,52 +178,33 @@ re_read:
 		BAD_ERROR(" Ambiguous sortlist entry \"%s\"\n\nIt maps to more than one source entry!  Please use an absolute path.\n", path);
 
 error:
-	fprintf(stderr, "Cannot stat sortlist entry \"%s\"\n", path);
-	fprintf(stderr, "This is probably because you're using the wrong file\n");
-	fprintf(stderr, "path relative to the source directories\n");
-	return FALSE;
+        fprintf(stderr, "Cannot stat sortlist entry \"%s\"\n", path);
+        fprintf(stderr, "This is probably because you're using the wrong file\n");
+        fprintf(stderr, "path relative to the source directories\n");
+        return FALSE;
 }
 
 
-void generate_file_priorities(char *pathname, int priority, struct stat *buf)
+void generate_file_priorities(struct dir_info *dir, int priority, struct stat *buf)
 {
-	char filename[8192];
-	DIR *linuxdir;
-	struct dirent *d_name;
-	
-	priority = get_priority(pathname, buf, priority);
+	priority = get_priority(dir->pathname, buf, priority);
 
-	if((linuxdir = opendir(pathname)) == NULL) {
-		ERROR("Could not open %s, skipping...\n", pathname);
-		return;
-	}
-	
-	while((d_name = readdir(linuxdir)) != NULL) {
-		if(strcmp(d_name->d_name, ".") == 0 || strcmp(d_name->d_name, "..") == 0)
-			continue;
-		strcat(strcat(strcpy(filename, pathname), "/"), d_name->d_name);
-
-		if(lstat(filename, buf) == -1) {
-			char buffer[8192];
-			sprintf(buffer, "Cannot stat dir/file %s, ignoring", filename);
-			perror(buffer);
-			continue;
-		}
-
-		if(excluded(filename, buf))
+	while(dir->current_count < dir->count) {
+		struct dir_ent *dir_ent = dir->list[dir->current_count++];
+		struct stat *buf = &dir_ent->inode->buf;
+		if(dir_ent->data)
 			continue;
 
 		switch(buf->st_mode & S_IFMT) {
 			case S_IFREG:
-				add_priority_list(filename, buf, get_priority(filename, buf, priority));
+				add_priority_list(dir_ent, get_priority(dir_ent->pathname, buf, priority));
 				break;
 			case S_IFDIR:
-				generate_file_priorities(filename, priority, buf);
+				generate_file_priorities(dir_ent->dir, priority, buf);
 				break;
 		}
 	}
-
-	closedir(linuxdir);
+	dir->current_count = 0;
 }
 
 
@@ -291,40 +228,30 @@ int read_sort_file(char *filename, int source, char *source_path[])
 }
 
 
-void sort_files_and_write(int source, char *source_path[])
+void sort_files_and_write(struct dir_info *dir)
 {
-	struct stat buf;
 	int i;
 	struct priority_entry *entry;
 	squashfs_inode inode;
 	int duplicate_file;
 
-	for(i = 0; i < source; i++) {
-		if(lstat(source_path[i], &buf) == -1) {
-			char buffer[8192];
-			sprintf(buffer, "Cannot stat dir/file %s, ignoring", source_path[i]);
-			perror(buffer);
-			continue;
-		}
-
-		if(excluded(source_path[i], &buf))
-			continue;
-
-		switch(buf.st_mode & S_IFMT) {
-			case S_IFREG:
-				add_priority_list(source_path[i], &buf, get_priority(source_path[i], &buf, 0));
-				break;
-			case S_IFDIR:
-				generate_file_priorities(source_path[i], 0, &buf);
-				break;
-		}
-	}
+	generate_file_priorities(dir, 0, &dir->dir_ent->inode->buf);
 
 	for(i = 65535; i >= 0; i--)
 		for(entry = priority_list[i]; entry; entry = entry->next) {
-			TRACE("%d: %s\n", i - 32768, entry->filename);
-			write_file(&inode, entry->filename, entry->size, &duplicate_file);
-			INFO("file %s, uncompressed size %lld bytes, %s\n", entry->filename, entry->size, duplicate_file ? "DUPLICATE" : "");
-			add_to_sorted_inode_list(inode, entry->st_dev, entry->st_ino);
+			TRACE("%d: %s\n", i - 32768, entry->dir->pathname);
+			if(entry->dir->inode->inode == SQUASHFS_INVALID_BLK) {
+				if(write_file(&inode, entry->dir, entry->dir->inode->buf.st_size,
+							&duplicate_file)) {
+					INFO("file %s, uncompressed size %lld bytes %s\n",
+						entry->dir->pathname,
+						entry->dir->inode->buf.st_size,
+						duplicate_file ? "DUPLICATE" : "");
+					entry->dir->inode->inode = inode;
+					entry->dir->inode->type = SQUASHFS_FILE_TYPE;
+				}
+			} else
+				INFO("file %s, uncompressed size %lld bytes LINK\n",
+					entry->dir->pathname, entry->dir->inode->buf.st_size);
 		}
 }
